from dataclasses import dataclass
from typing import Literal


@dataclass
class CurrentState:
    task_id: int | None = None
    quiz_id: str | None = None
    app_title: str = "Клондайк аналитика"
    app_subtitle: str = ""
    # название выбранного объекта
    object_name: Literal["quiz", "task", "account"] | None = None
    # file full path
    opened_file_path: str = ""
    # just file name
    opened_file_name: str = ""
    # just opened file min required version
    opened_file_min_supported_app_version: str = ""
    app_version: str = ""



from analyst_klondike.features.code import code_reducer
from analyst_klondike.features.code_explorer import code_explorer_reducer
from analyst_klondike.features.data_context import load_file_reducer
from analyst_klondike.features.app import app_reducer
from analyst_klondike.features.message_box import message_box_reducer
from analyst_klondike.state.app_state import dispatch
from analyst_klondike.state.base_action import BaseAction


def app_dispatch(action: BaseAction):
    dispatch(action,
             load_file_reducer.apply,
             code_explorer_reducer.apply,
             code_reducer.apply,
             app_reducer.apply,
             message_box_reducer.apply
             )



import os
from textual import on
from textual.reactive import var
from textual.app import ComposeResult
from textual.screen import ModalScreen
from textual.containers import Grid, Horizontal
from textual.widgets import Button, Label, DirectoryTree


class OpenFileScreen(ModalScreen[str]):

    CSS_PATH = "open_file_screen.tcss"
    selected_path = var("")

    def on_mount(self) -> None:
        open_button = self.query_one("#open", Button)
        open_button.disabled = True

    def compose(self) -> ComposeResult:
        with Grid(id="dialog"):
            yield Label("Выберите файл", id="modal_question")
            yield DirectoryTree("./")
            with Horizontal():
                yield Button("Открыть", id="open", variant="primary")
                yield Button("Отмена", id="cancel", variant="error")

    @on(DirectoryTree.FileSelected)
    def on_file_selected(self, ev: DirectoryTree.FileSelected) -> None:
        self.selected_path = str(ev.path)
        open_button = self.query_one("#open", Button)
        open_button.disabled = not os.path.exists(self.selected_path)

    @on(Button.Pressed, "#open")
    def on_ok_button_pressed(self) -> None:
        self.dismiss(self.selected_path)

    @on(Button.Pressed, "#cancel")
    def on_cancel_button_pressed(self) -> None:
        self.dismiss(None)



from analyst_klondike.features.message_box.actions import (
    DisplayMessageBoxAction,
    HideMessageBoxAction
)
from analyst_klondike.state.app_state import AppState
from analyst_klondike.state.base_action import BaseAction


def apply(state: AppState, action: BaseAction) -> AppState:
    if isinstance(action, DisplayMessageBoxAction):
        state.message_box.is_visible = True
        state.message_box.message = action.message
        state.message_box.ok_button_callback = action.ok_button_callback
        return state
    if isinstance(action, HideMessageBoxAction):
        state.message_box.is_visible = False
        return state
    return state



from __future__ import annotations
from os.path import basename
from typing import Any

from textual import on
from textual.message import Message
from textual.app import ComposeResult
from textual.screen import Screen
from textual.containers import Vertical
from textual.widgets import (
    Header,
    Footer
)
from textual.binding import Binding

# message box
from analyst_klondike.features.app.open_file_action import open_file
from analyst_klondike.features.code.selector import has_failed_cases
from analyst_klondike.features.data_context.data_state import PythonTaskState
from analyst_klondike.features.events.event_hook import send_event
from analyst_klondike.features.message_box.ui.mb_screen import MessageBoxScreen

# dispatch and state
from analyst_klondike.features.data_context.set_opened_file_action import SetOpenedFileAction
from analyst_klondike.features.update.display_has_later_version import display_message_if_outdated
from analyst_klondike.state.app_dispatch import app_dispatch
from analyst_klondike.state.app_state import (
    AppState,
    get_state,
    select)

# selectors
from analyst_klondike.features.current.selectors import select_current_task, select_has_file_openned

# actions
from analyst_klondike.features.app.actions import EditorScreenReadyAction
from analyst_klondike.features.code.actions import (
    RunCodeAndSetResultsAction,
    UpdateCodeAction)
from analyst_klondike.features.current.actions import (
    MakeQuizCurrentAction,
    MakeTaskCurrentAction
)
from analyst_klondike.features.code_explorer.code_explorer_reducer import (
    QuizNodeCollapseAction,
    QuizNodeExpandAction
)
from analyst_klondike.features.data_context.save_action import save_to_yaml

from analyst_klondike.features.message_box.actions import (
    DisplayMessageBoxAction,
    HideMessageBoxAction)


# components
from analyst_klondike.state.selectors import select_task_by_id
from analyst_klondike.ui.editor_screen.components.code_editor import CodeEditor
from analyst_klondike.ui.editor_screen.components.current_task import CurrentTaskInfo
from analyst_klondike.ui.editor_screen.components.explorer import Explorer
from analyst_klondike.ui.editor_screen.components.quiz_description import QuizDescription
from analyst_klondike.ui.editor_screen.components.test_results import TestResults
from analyst_klondike.ui.file_screen.open_file_screen import OpenFileScreen

# other
from analyst_klondike.features.wellcome_message import get_wellcome_message_text


_WELLCOME_MESSAGE = get_wellcome_message_text()


class EditorScreen(Screen[Any]):
    CSS_PATH = "editor.tcss"

    open_file_binding = Binding(id="open_file",
                                key='ctrl+o',
                                action='open_quiz_file',
                                description="Открыть тест",
                                tooltip="Открыть файл с задачами. " +
                                "Вам нужно будет написать код, который пройдет все тесты")

    run_code_binding = Binding(id="run_code",
                               key='f5, ctrl+r',
                               action='run_btn_click',
                               description="Запустить код")

    save_file_binding = Binding(id="save_file",
                                key="ctrl+s",
                                action="save_quiz_to_file",
                                description="Сохранить")

    display_help_binding = Binding(id="display_help",
                                   key="f1",
                                   action="display_help",
                                   description="Справка",
                                   tooltip="Отобразить справку")

    MSG_BOX_SCREEN_ID = "message_box_screen"

    BINDINGS = [
        display_help_binding,
        open_file_binding,
        save_file_binding,
        run_code_binding
    ]

    class UpdateAppTitleMessage(Message):
        def __init__(self, title: str, subtitle: str) -> None:
            super().__init__()
            self.title = title
            self.subtitle = subtitle

    class RequestOpenFileScreen(Message):
        pass

    def _notify_opened(self, fname: str | None) -> None:
        if fname is None:
            return
        self.app.notify(
            "Загружено",
            title=fname,
            severity="information",
            timeout=1
        )

    async def on_mount(self) -> None:
        app_dispatch(EditorScreenReadyAction())
        app_dispatch(DisplayMessageBoxAction(_WELLCOME_MESSAGE))
        open_file(self._notify_opened)
        send_event("app_opened", "/")
        display_message_if_outdated()

    def update_view(self, new_state: AppState):
        if not new_state.is_editor_screen_ready:
            return
        explorer = self.query_one("Explorer", Explorer)
        code_editor = self.query_one("CodeEditor", CodeEditor)
        task_info = self.query_one("CurrentTaskInfo", CurrentTaskInfo)
        test_results = self.query_one("TestResults", TestResults)
        quiz_description = self.query_one("QuizDescription", QuizDescription)

        explorer.state = new_state
        code_editor.state = new_state
        task_info.state = new_state
        test_results.update_view(new_state)
        quiz_description.state = new_state

        # send message to update title and subtitle
        self.post_message(EditorScreen.UpdateAppTitleMessage(
            new_state.current.app_title,
            new_state.current.app_subtitle
        ))
        # update component visibility
        if new_state.current.object_name == "task":
            code_editor.remove_class("component-hidden")
            quiz_description.add_class("component-hidden")
        elif new_state.current.object_name == "quiz":
            code_editor.add_class("component-hidden")
            quiz_description.remove_class("component-hidden")
        elif new_state.current.object_name == "account":
            pass

        self._display_message_box(new_state)

    def _display_message_box(self, state: AppState) -> None:
        current_screen_id = self.app.screen_stack[-1].id

        # display message box:
        if state.message_box.is_visible:
            if current_screen_id != EditorScreen.MSG_BOX_SCREEN_ID:
                self.app.push_screen(
                    MessageBoxScreen(
                        id=EditorScreen.MSG_BOX_SCREEN_ID),
                    EditorScreen._on_message_box_close
                )
        else:
            if current_screen_id == EditorScreen.MSG_BOX_SCREEN_ID:
                self.app.pop_screen()

    def compose(self) -> ComposeResult:
        yield Header()
        yield Footer()
        with Vertical(id="left_panel"):
            yield Explorer()
            yield CurrentTaskInfo()
        with Vertical(id="right_panel"):
            yield CodeEditor()
            yield QuizDescription(classes="component-hidden")
            yield TestResults()

    @on(Explorer.TaskSelected)
    def on_python_task_selected(self, ev: Explorer.TaskSelected) -> None:
        if isinstance(ev.task_id, int):
            app_dispatch(MakeTaskCurrentAction(ev.task_id))
            selected_task = select_task_by_id(get_state(), ev.task_id)
            send_event("python_task_selected", "/",
                       event_json={
                           'id': str(ev.task_id),
                           'title': selected_task.title
                       })

    @on(Explorer.QuizSelected)
    def on_quiz_selected(self, ev: Explorer.QuizSelected) -> None:
        app_dispatch(MakeQuizCurrentAction(quiz_id=ev.quiz_id))

    @on(Explorer.QuizNodeExpandedOrCollapsed)
    def on_quiz_node_expanded_collapsed(self, ev: Explorer.QuizNodeExpandedOrCollapsed) -> None:
        if ev.action_type == "expanded":
            app_dispatch(QuizNodeExpandAction(quiz_id=ev.quiz_id))
        elif ev.action_type == "collapsed":
            app_dispatch(QuizNodeCollapseAction(quiz_id=ev.quiz_id))

    @on(CodeEditor.CodeUpdated)
    def on_editor_code_updated(self, ev: CodeEditor.CodeUpdated) -> None:
        app_dispatch(UpdateCodeAction(ev.code))

    def action_run_btn_click(self) -> None:
        app_dispatch(RunCodeAndSetResultsAction())
        is_failed = select(has_failed_cases)
        selected_task: PythonTaskState = select(select_current_task)
        send_event("run_code", "/", event_json={
            'is_passed': not is_failed,
            'task_id': selected_task.id,
            'task_title': selected_task.title
        })

    def action_open_quiz_file(self) -> None:
        def _on_file_selected(file_path: str | None) -> None:
            if not isinstance(self.app.screen, EditorScreen):
                return
            if file_path is None or file_path == '':
                return

            app_dispatch(SetOpenedFileAction(
                opened_file_name=basename(file_path),
                opened_file_path=file_path
            ))
            open_file(self._notify_opened)

        self.app.push_screen(OpenFileScreen(), _on_file_selected)

    def action_save_quiz_to_file(self) -> None:
        state = get_state()
        save_to_yaml(state, self.app)

    def action_display_help(self) -> None:
        app_dispatch(DisplayMessageBoxAction(_WELLCOME_MESSAGE))

    def check_action(self, action: str, parameters: tuple[object, ...]):
        if action == EditorScreen.run_code_binding.action:
            has_file = select(select_has_file_openned)
            # Кнопка должна быть активна, если задача активирована
            if has_file:
                return True
            return None
        if action == EditorScreen.save_file_binding.action:
            has_file = select(select_has_file_openned)
            if has_file:
                return True
            return None
        return True

    @staticmethod
    def _on_message_box_close(_: bool | None) -> None:
        app_dispatch(HideMessageBoxAction())



import asyncio
from typing import Any, Literal
from analyst_klondike.features.events.api_sender import ApiEventSender
from analyst_klondike.features.events.event_sender import BaseEventSender
from analyst_klondike.features.events.file_sender import FileEventSender
from analyst_klondike.state.app_state import select

from analyst_klondike.state.app_state import AppState


EventSenderType = Literal["local_file", "local_db", "remote_db"]


def create_event_sender(sender_type: EventSenderType) -> BaseEventSender:
    if sender_type == "local_file":
        return FileEventSender()
    if sender_type == "local_db":
        return ApiEventSender(
            api_url="https://localhost:7150/api/NeonApi/PostEvent")
    if sender_type == "remote_db":
        return ApiEventSender(
            api_url="https://www.xn----7sbbaat6aaehdfdhwgj7f.xn--p1ai/api/NeonApi/PostEvent")
    raise ValueError(f"Event sender <{sender_type}> supported")


def send_event(event_name: str,
               event_location: str,
               event_json: dict[str, str | int | bool] | None = None,
               async_run: bool = True) -> None:
    # pylint: disable=import-outside-toplevel
    from analyst_klondike.ui.runner_app import get_app

    async def _send_async(event_name: str,
                          event_location: str) -> None:
        # await asyncio.sleep(10)
        event_sender = _use_event_sender()
        await event_sender.send_event(event_name=event_name,
                                      event_location=event_location,
                                      event_json=event_json)

    app = get_app()
    if async_run:
        app.run_worker(_send_async(event_name, event_location), thread=True)
    else:
        asyncio.create_task(_send_async(event_name, event_location))


def event(event_name: str,
          event_location: str,
          event_start_prefix: str | None = None,
          event_end_prefix: str | None = None) -> Any:

    def actual_decorator(func: Any):

        def wrapper(*args: Any, **kwargs: Any):
            if event_start_prefix is not None and event_end_prefix is not None:
                send_event(event_name + "_start", event_location)
                func(*args, **kwargs)
                send_event(event_name + "_end", event_location)

            if event_end_prefix is None and event_start_prefix is None:
                func(*args, **kwargs)
                send_event(event_name, event_location)
        return wrapper

    return actual_decorator


def _use_event_sender() -> BaseEventSender:
    sender_type = select(_select_senver)
    return create_event_sender(sender_type)


def _select_senver(state: AppState) -> str:
    return state.event_sender



from os.path import dirname, join, realpath
from textwrap import dedent


def _fpath() -> str:
    this_file_dir = dirname(realpath(__file__))
    return join(this_file_dir, "message_text.txt")


def get_wellcome_message_text():
    with open(_fpath(), encoding='UTF-8') as f:
        return dedent(f.read())



from dataclasses import dataclass, field
from typing import Literal

PythonTaskResult = Literal["passed", "failed", "not_runned"]


@dataclass
class VariableState:
    param_name: str
    param_type: str
    param_value: str


@dataclass
class TestCaseState:
    inputs: list[VariableState]
    expected: VariableState


@dataclass
class PythonTaskState:
    id: int
    title: str
    description: str
    code_template: str
    code: str
    quiz_id: str
    test_cases: list[TestCaseState] = field(
        default_factory=list[TestCaseState]
    )
    is_passed: PythonTaskResult = "not_runned"


@dataclass
class PythonQuizState:
    id: str
    title: str
    is_node_expanded: bool


@dataclass
class DataState:
    quizes: dict[str, PythonQuizState] = field(
        default_factory=dict[str, PythonQuizState]
    )
    tasks: dict[int, PythonTaskState] = field(
        default_factory=dict[int, PythonTaskState]
    )



from dataclasses import dataclass
from analyst_klondike.state.base_action import BaseAction


@dataclass
class ChangeThemeAction(BaseAction):
    type = "CHANGE_THEME_ACTION"
    theme: str
    is_dark: bool


@dataclass
class EditorScreenReadyAction(BaseAction):
    type = "EDITOR_SCREEN_READY_ACTION"



from datetime import datetime
from analyst_klondike.features.code.actions import (
    RunAllCodeAction,
    RunCodeAndSetResultsAction,
    UpdateCodeAction
)
from analyst_klondike.features.code.code_runner.runner import CodeRunner
from analyst_klondike.features.code.state import ResultCaseState
from analyst_klondike.features.data_context.data_state import PythonTaskState
from analyst_klondike.state.app_state import AppState
from analyst_klondike.state.base_action import BaseAction
from analyst_klondike.features.current.selectors import select_current_task


def apply(state: AppState, action: BaseAction) -> AppState:
    if isinstance(action, UpdateCodeAction):
        # update current code
        state.run_code.last_code = action.code

        # update current task code
        current_task = select_current_task(state)
        if current_task is not None:
            current_task.code = action.code

        # set code to run
        state.run_code.last_code = action.code
        return state
    if isinstance(action, RunCodeAndSetResultsAction):
        _run_and_set_results(state)
        return state
    if isinstance(action, RunAllCodeAction):
        _run_previously_runned_tasks(state)
    return state


def _run_and_set_results(state: AppState) -> None:
    current_task = select_current_task(state)
    if current_task is None:
        return
    _run_task(current_task, state)


def _run_previously_runned_tasks(state: AppState) -> None:
    runned_tasks = (t for t in state.data.tasks.values()
                    if t.is_passed in ("passed", "failed"))
    for task in runned_tasks:
        _run_task(task, state)


def _run_task(current_task: PythonTaskState, state: AppState) -> None:
    # run code here and set results
    res = CodeRunner(function_name="solution").run_code(
        current_task.test_cases,
        current_task.code)

    # set last code run info
    state.run_code.last_run_dt = datetime.now()
    state.run_code.last_code = current_task.code

    state.run_code.result_cases = [
        ResultCaseState(
            func_params=r.func_params,
            expected=r.expected,
            actual=r.actual,
            passed=r.passed,
            has_exception=r.exception is not None,
            exception_message=r.exception,
            printed_lines=r.printed_lines
        ) for r in res.all_cases
    ]
    # add test case errors
    state.run_code.errors = list(res.errors)
    state.run_code.has_errors = len(res.errors) > 0

    are_all_tests_passed = all(c.passed for c in res.all_cases)

    is_task_passed = not state.run_code.has_errors and are_all_tests_passed

    # set run result:
    current_task.is_passed = "passed" if is_task_passed else "failed"



from textwrap import dedent
from rich.text import Text
from textual import on
from textual.reactive import reactive
from textual.app import ComposeResult
from textual.containers import (
    Vertical,
    Horizontal,
    Container,
    VerticalScroll)
from textual.widget import Widget
from textual.widgets import Button, Static
from textual.screen import ModalScreen
from analyst_klondike.state.app_state import AppState, get_state, select


class MessageText(Widget):

    markup_text = reactive("")

    def __init__(self, text: str) -> None:
        super().__init__()
        self.markup_text = text

    def render(self) -> Text:
        return Text.from_markup(self.markup_text)


class MessageBoxScreen(ModalScreen[bool]):

    @staticmethod
    def message(state: AppState) -> str:
        return dedent(state.message_box.message)

    CSS_PATH = "mb_screen.tcss"

    def compose(self) -> ComposeResult:
        msg_text = select(MessageBoxScreen.message)
        yield Container(id="shadow")
        with Vertical(id="dialog"):
            with VerticalScroll():
                yield Static(msg_text)
            with Horizontal(id="buttons"):
                yield Button(id="ok_button",
                             label="ОК",
                             variant="success")

    @on(Button.Pressed, "#ok_button")
    def ok_button_click(self) -> None:
        state = get_state()
        if state.message_box.ok_button_callback is not None:
            state.message_box.ok_button_callback()
        self.dismiss(True)



import sys
from analyst_klondike.features.data_context.arg_file_load.arg_load import (
    TooManyArgumentsException,
    set_file_from_argv
)
from analyst_klondike.ui.runner_app import get_app


def analyst_klondike():
    try:
        app = get_app()
        set_file_from_argv(sys.argv)
        app.run()
    except (FileNotFoundError, TooManyArgumentsException) as err:
        print(err)
        sys.exit(1)


if __name__ == "__main__":
    analyst_klondike()



from typing import Any
from textual import on
from textual.app import App
import pyperclip


from analyst_klondike.features.current.selectors import select_has_file_openned
from analyst_klondike.features.data_context.save_action import save_to_yaml
from analyst_klondike.features.app.actions import ChangeThemeAction
from analyst_klondike.features.demo_quiz.demo import start_demo_quiz
from analyst_klondike.features.events.event_hook import send_event
from analyst_klondike.state.app_dispatch import app_dispatch
from analyst_klondike.state.app_state import AppState, get_state, select
from analyst_klondike.ui.editor_screen.editor import EditorScreen
from analyst_klondike.ui.file_screen.save_file_before_exit_screen import (
    SaveOnExitModal,
    SaveOrExitModalResult
)


class RunnerApp(App[Any]):
    COMMAND_PALETTE_BINDING = "ctrl+backslash"
    EDITOR_SCREEN = EditorScreen()

    def __init__(self) -> None:
        super().__init__()

    def update_view(self, state: AppState) -> None:
        RunnerApp.EDITOR_SCREEN.update_view(state)

    def on_mount(self) -> None:
        self.push_screen(RunnerApp.EDITOR_SCREEN)
        self.theme = "gruvbox"
        self.title = "Клондайк аналитика"
        self.sub_title = "Интерактивный тренажер Python на вашем компьютере"

    @on(EditorScreen.UpdateAppTitleMessage)
    def on_title_subtitle_changed(self, event: EditorScreen.UpdateAppTitleMessage) -> None:
        self.title = event.title
        self.sub_title = event.subtitle

    async def action_quit(self):

        def on_screen_get_result(res: SaveOrExitModalResult | None) -> None:
            if res == "cancel":
                return
            if res == "save":
                state = get_state()
                save_to_yaml(state, self.app)
            send_event("app_normally_closed", "/")
            self.app.exit(0)
        has_file = select(select_has_file_openned)
        if has_file:
            self.push_screen(SaveOnExitModal(), on_screen_get_result)
        else:
            send_event("app_normally_closed", "/", async_run=False)
            self.exit(0)

    def watch_theme(self, new_theme: Any) -> None:
        app_dispatch(ChangeThemeAction(
            theme=new_theme,
            is_dark=self.current_theme.dark
        ))

    # This action is used in wellcome message
    def action_start_demo_test(self) -> None:
        start_demo_quiz()

    # Copy to clipboard update string
    def action_copy_update_str_to_clipboard(self) -> None:
        try:
            pyperclip.copy('uv tool upgrade analyst-klondike')
            self.notify(
                message="Скопировано uv tool upgrade analyst-klondike",
                title="Клондайк аналитика",
                severity="information"
            )
            send_event("update_command_is_copied", "/")
        except Exception:  # pylint: disable=W0718
            self.notify(
                message="Ошибка при копировании uv tool upgrade analyst-klondike в буфер обмена",
                title="Клондайк аналитика",
                severity="error"
            )


_app = RunnerApp()


def get_app() -> RunnerApp:
    return _app



from copy import deepcopy
from dataclasses import dataclass
from typing import Any, Callable, Literal
from analyst_klondike.features.code.state import RunCodeState
from analyst_klondike.features.current.current_state import CurrentState
from analyst_klondike.features.data_context.data_state import DataState
from analyst_klondike.features.message_box.state import MessageBoxState
from analyst_klondike.state.base_action import BaseAction


@dataclass
class Enable:
    run_code_button: bool
    open_file_button: bool


@dataclass
class AppState:
    last_action_type: str
    user_email: str
    theme: str
    is_dark: bool
    is_editor_screen_ready: bool
    event_sender: Literal["local_file", "local_db", "remote_db"]
    data: DataState
    current: CurrentState
    run_code: RunCodeState
    message_box: MessageBoxState


INITIAL_STATE = AppState(
    user_email="",
    theme="",
    is_dark=True,
    is_editor_screen_ready=False,
    data=DataState(),
    current=CurrentState(),
    run_code=RunCodeState(),
    message_box=MessageBoxState(),
    last_action_type="",
    event_sender="remote_db"  # "local_file"
)


_state = INITIAL_STATE


def get_state() -> AppState:
    return select(lambda _: _state)


Reducers = Callable[[AppState, BaseAction], AppState]


def dispatch(action: BaseAction, *reducers: Reducers) -> None:
    state = get_state()
    copy_state = deepcopy(state)
    copy_state.last_action_type = action.type
    # pylint: disable=global-statement
    global _state
    _state = _pipe(copy_state, action, *reducers)
    # pylint: disable=import-outside-toplevel
    from analyst_klondike.ui.runner_app import get_app
    app = get_app()
    app.update_view(_state)


def select(selector: Callable[[AppState], Any]) -> Any:
    return selector(_state)


def _pipe(state: AppState,
          action: BaseAction,
          *reducers: Callable[[AppState, BaseAction], AppState]) -> AppState:
    s = state
    for apply in reducers:
        s = apply(s, action)
    return s



from dataclasses import dataclass
import analyst_klondike
from analyst_klondike.features.data_context.data_state import (
    PythonQuizState,
    PythonTaskResult,
    PythonTaskState,
    TestCaseState,
    VariableState
)
from analyst_klondike.features.data_context.json_load.dc import (
    JsonLoadResult,
    PythonQuizJson,
    QuestionJson,
    TestCaseJson)
from analyst_klondike.state.app_state import AppState
from analyst_klondike.state.base_action import BaseAction


@dataclass
class InitAction(BaseAction):
    type = "INIT_APP"
    data: JsonLoadResult


@dataclass
class ReadFileAndAppVersion(BaseAction):
    type = "INIT_FILE_AND_APP_VERSION"
    data: JsonLoadResult


def init_state(state: AppState, data: JsonLoadResult) -> None:
    state.user_email = "Клондайк аналитика"
    state.data.quizes = {
        x.id: _get_quiz_state(x) for x in data.quizes
    }
    state.data.tasks = {
        question.id: _get_question_state(question, quiz.id)
        for quiz in data.quizes
        for question in quiz.questions
    }
    state.current.app_title = "Клондайк аналитика"
    state.current.app_subtitle = "Интерактивный тренажер Python на вашем " +\
        f"компьютере (version. {analyst_klondike.__version__})"
    state.current.app_version = analyst_klondike.__version__
    state.current.opened_file_min_supported_app_version = \
        data.quiz_info.min_supported_app_version


def init_file_and_app_version(state: AppState, data: JsonLoadResult) -> None:
    state.current.app_version = analyst_klondike.__version__
    state.current.opened_file_min_supported_app_version = \
        data.quiz_info.min_supported_app_version


def _get_quiz_state(data: PythonQuizJson) -> PythonQuizState:
    return PythonQuizState(
        id=data.id,
        title=data.title,
        is_node_expanded=False
    )


def _get_question_state(data: QuestionJson, quiz_id: str) -> PythonTaskState:
    def get_passed_status() -> PythonTaskResult:
        if data.is_passed == "passed":
            return "passed"
        if data.is_passed == "failed":
            return "failed"
        return "not_runned"

    return PythonTaskState(
        id=data.id,
        quiz_id=quiz_id,
        title=data.title,
        description=data.text,
        code_template=data.code_template,
        code=data.code if data.code != "" else data.code_template,
        test_cases=_get_test_cases(data.test_cases),
        is_passed=get_passed_status(),
    )


def _get_test_cases(cases: list[TestCaseJson]) -> list[TestCaseState]:
    return [
        TestCaseState(
            inputs=[VariableState(
                param_name=inp.param_name,
                param_type=inp.param_type,
                param_value=inp.param_value
            ) for inp in c.inputs],
            expected=VariableState(
                param_name=c.expected.param_name,
                param_value=c.expected.param_value,
                param_type=c.expected.param_type
            )
        )
        for c in cases
    ]



# pylint: disable=W0122

from types import CodeType
from copy import copy
import textwrap
from typing import Any, Callable, Iterator

from analyst_klondike.features.code.code_runner.result_case import ResultCase, RunResult
from analyst_klondike.features.data_context.data_state import TestCaseState, VariableState


class CodeRunner:

    class SolutionFuncError(Exception):
        pass

    def __init__(self, function_name: str = "solution") -> None:
        self._function_name = function_name
        self._printed_lines: list[str] = []

    def run_code(self,
                 cases: list[TestCaseState],
                 code: str) -> RunResult:
        printed_lines: list[Any] = []

        cleaned_code = textwrap.dedent(code)
        try:
            compiled_code = CodeRunner._try_compile(cleaned_code,)
            function_to_run = self._try_get_solution_func(
                compiled_code
            )
        except (SyntaxError, NameError) as synt_err:
            return RunResult.create_error_result([str(synt_err)])
        except CodeRunner.SolutionFuncError as func_err:
            return RunResult.create_error_result([str(func_err)])

        result_cases = list(self._run_test_cases(cases, function_to_run))
        distinct_errors = list(
            set(c.exception for c in result_cases if c.exception is not None))

        return RunResult(
            all_cases=result_cases,
            printed_lines=printed_lines,
            errors=distinct_errors)

    def _try_get_solution_func(self,
                               compiled_code: CodeType | None
                               ) -> Callable[..., Any]:
        if compiled_code is None:
            raise ValueError("No code to run")
        loc: dict[str, object] = {}
        exec_globals = copy(globals())
        exec_globals.update({
            'print': self._print
        })
        exec(compiled_code, exec_globals, loc)
        if self._function_name not in loc:
            raise CodeRunner.SolutionFuncError(
                f"No function {self._function_name} in code. You may have accidentally deleted it.")
        function_to_run = loc[self._function_name]
        if not callable(function_to_run):
            raise CodeRunner.SolutionFuncError(
                f"Function {self._function_name} not recognized")
        return function_to_run

    @staticmethod
    def _try_compile(code: str) -> CodeType:
        compile_result = compile(code, '<string>', "exec")
        return compile_result

    def _run_test_cases(self,
                        cases: list[TestCaseState],
                        func: Callable[..., Any]) -> Iterator[ResultCase]:
        for c in cases:
            params = dict(
                (inp.param_name, self._cast(inp))
                for inp in c.inputs if inp.param_name != "expected"
            )

            exp = self._cast(c.expected)
            try:
                self._printed_lines.clear()
                actual = func(**params)
                result_case = ResultCase(
                    func_params=params,
                    expected=exp,
                    actual=actual,
                    passed=str(exp) == str(actual),
                    printed_lines=self._printed_lines[:]
                )
                self._printed_lines.clear()
                yield result_case
            except Exception as ex:
                yield ResultCase(
                    func_params=params,
                    expected=exp,
                    actual=None,
                    passed=False,
                    exception=str(ex),
                    printed_lines=[]
                )

    def _print(self, *args: Any) -> None:
        self._printed_lines.append(" ".join(str(a) for a in args))

    def _cast(self, p: VariableState) -> Any:
        if p.param_type == "int":
            return int(p.param_value)
        if p.param_type == "float":
            return float(p.param_value)
        if p.param_type == "list":
            return p.param_value
        return str(p.param_value)



from textual.app import ComposeResult
from textual.reactive import var
from textual.containers import Vertical
from textual.widgets import Static

from analyst_klondike.state.app_state import INITIAL_STATE, AppState
from analyst_klondike.features.current.selectors import select_current_quiz


class QuizDescription(Vertical):

    state = var(INITIAL_STATE, init=False)

    DEFAULT_CSS = """
        QuizDescription {
            border: solid $primary;
            padding: 1;
        }
    """

    def on_mount(self) -> None:
        self.border_title = "Тест"

    def compose(self) -> ComposeResult:
        yield Static("Hello, quiz")

    def watch_state(self, new_state: AppState):
        static = self.query_one("Static", Static)
        current_quiz = select_current_quiz(new_state)
        if current_quiz is not None:
            self.border_title = current_quiz.title
            static.update(current_quiz.title)



from textual import on
from textual.message import Message
from textual.app import ComposeResult
from textual.widgets import TextArea
from textual.containers import VerticalScroll
from textual.reactive import var

from analyst_klondike.features.current.actions import MakeTaskCurrentAction
from analyst_klondike.features.app.actions import ChangeThemeAction
from analyst_klondike.state.app_state import INITIAL_STATE, AppState
from analyst_klondike.features.current.selectors import select_current_task


class CodeEditor(VerticalScroll):

    DEFAULT_CSS = """
        CodeEditor {
            border: solid $primary;

            TextArea {
                border: none;
                
                &:focus-within {
                    border: none;
                }
            }
        }
    """

    class CodeUpdated(Message):
        def __init__(self, code: str) -> None:
            super().__init__()
            self.code = code

    state: var[AppState] = var(INITIAL_STATE, init=False)

    def on_mount(self):
        self.border_title = "Редактор"
        textarea = self.query_one("TextArea", TextArea)
        textarea.indent_type = "tabs"

    def compose(self) -> ComposeResult:
        code_editor = TextArea.code_editor("",
                                           theme="monokai",
                                           language="python",
                                           id="code_editor")
        yield code_editor

    def watch_state(self, state: AppState) -> None:
        if state.last_action_type in (MakeTaskCurrentAction.type):
            task = select_current_task(state)
            editor = self.query_one("TextArea", TextArea)
            if task is None:
                return
            with editor.prevent(TextArea.Changed):
                editor.text = task.code
        if state.last_action_type == ChangeThemeAction.type:
            editor = self.query_one("TextArea", TextArea)
            if state.is_dark:
                editor.theme = 'monokai'
            else:
                editor.theme = 'github_light'

    @on(TextArea.Changed)
    def on_textarea_changed(self, event: TextArea.Changed) -> None:
        new_code = event.text_area.text
        self.post_message(CodeEditor.CodeUpdated(new_code))
